@@ -2,6 +2,8 @@
     # Workaround / say if using cached Charm present
     # Write out resolved deployment file with includes & overrides
     # Ensure bootstrap
+
+    # Deployment validate
 """
 
 from base64 import b64encode
@@ -96,11 +98,9 @@ loggers:
 """
 
 
-def setup_logging(verbose=False, debug=False):
+def setup_logging(verbose=False, debug=False, stream=None):
     config = yaml_load(DEFAULT_LOGGING)
     log_options = {}
-    if debug and not verbose:
-        verbose = True
     if verbose:
         log_options.update({"loggers": {"": {"level": "DEBUG"}}})
     if debug:
@@ -108,6 +108,14 @@ def setup_logging(verbose=False, debug=False):
     config = dict_merge(config, log_options)
     logConfig(config)
 
+    # Allow tests to reuse this func to mass configure log streams.
+    if stream:
+        root = logging.getLogger()
+        previous = root.handlers[0]
+        root.handlers[0] = current = logging.StreamHandler(stream)
+        current.setFormatter(previous.formatter)
+        return stream
+
 
 @contextmanager
 def temp_file():
@@ -137,6 +145,10 @@ def _check_call(params, cwd, log, *args, **kw):
 # Utils from deployer 1
 def relations_combine(onto, source):
     target = deepcopy(onto)
+    # Support list of relations targets
+    if isinstance(onto, list) and isinstance(source, list):
+        target.extend(source)
+        return target
     for (key, value) in source.items():
         if key in target:
             if isinstance(target[key], dict) and isinstance(value, dict):
@@ -151,12 +163,11 @@ def relations_combine(onto, source):
 def dict_merge(onto, source):
     target = deepcopy(onto)
     for (key, value) in source.items():
-        if (key in target and isinstance(target[key], dict) and
+        if key == 'relations' and key in target:
+            target[key] = relations_combine(target[key], value)
+        elif (key in target and isinstance(target[key], dict) and
                 isinstance(value, dict)):
-            if key == 'relations':
-                target[key] = relations_combine(target[key], value)
-            else:
-                target[key] = dict_merge(target[key], value)
+            target[key] = dict_merge(target[key], value)
         else:
             target[key] = value
     return target
@@ -184,7 +195,7 @@ def setup_parser():
                       dest="debug",
                       action="store_true", default=False)
     parser.add_option('-L', '--local-mods',
-                      help='Allow deployment of locally-modified charms',
+                      help='Disallow deployment of locally-modified charms',
                       dest="no_local_mods", default=True, action='store_false')
     parser.add_option('-u', '--update-charms',
                       help='Update existing charm branches',
@@ -206,14 +217,13 @@ def setup_parser():
     parser.add_option("-f", '--find-service', action="store", type="string",
                       help='Find hostname from first unit of a specific service.',
                       dest="find_service")
-#    parser.add_option("-m", '--max-concurrent', action="store", type="int",
-#                      help=("Maximum number of concurrent deployments to send "
-#                            " to provider. Default: no limit"),
-#                      dest="max_concur_deploy", default=0)
+    parser.add_option("-b", '--branch-only', action="store_true",
+                      help='Update vcs branches and exit.',
+                      dest="branch_only")
     parser.add_option('-s', '--deploy-delay', action='store', type='float',
                       help=("Time in seconds to sleep between 'deploy' commands, "
                             "to allow machine provider to process requests. This "
-                            "delay is also enforced between calls to"
+                            "delay is also enforced between calls to "
                             "terminate_machine"),
                       dest="deploy_delay", default=0)
     parser.add_option('-e', '--environment', action='store', dest='juju_env',
@@ -223,14 +233,13 @@ def setup_parser():
                       help=('Override *all* config options of the same name '
                             'across all services.  Input as key=value.'),
                       dest='overrides', default=None)
+    parser.add_option('-v', '--verbose', action='store_true', default=False,
+                      dest="verbose", help='Verbose output')
     parser.add_option('-W', '--watch', help='Watch environment changes on console',
                       dest="watch",
                       action="store_true", default=False)
-
-# bzr up -r doesn't seem to support revisionspec date forms??
-#    parser.add_option('-r', '--revdate', help='Deploy charm revisions as of date',
-#                      dest="revspec",
-#                      action="store", default=False)
+    parser.add_option('-r', "--retry", default=0, type=int, dest="retry_count",
+                      help="Resolve unit errors via retry. Either standalone or in a deployment")
     parser.add_option('-w', '--relation-wait', action='store', dest='rel_wait',
                       default=60, type=int,
                       help=('Number of seconds to wait before checking for '
@@ -300,6 +309,17 @@ class ConfigStack(object):
         return deploy_data
 
 
+class Endpoint(object):
+
+    def __init__(self, ep):
+        self.ep = ep
+        self.name = None
+        if ":" in self.ep:
+            self.service, name = self.ep.split(":")
+        else:
+            self.service = ep
+
+
 class Service(object):
 
     def __init__(self, name, svc_data):
@@ -318,6 +338,14 @@ class Service(object):
     def num_units(self):
         return int(self.svc_data.get('num_units', 1))
 
+    @property
+    def force_machine(self):
+        return self.svc_data.get('force-machine')
+
+    @property
+    def expose(self):
+        return self.svc_data.get('expose', False)
+
 
 class Vcs(object):
 
@@ -339,7 +367,7 @@ class Vcs(object):
             output = subprocess.check_output(
                 args, cwd=cwd or self.path, stderr=subprocess.STDOUT)
         except subprocess.CalledProcessError, e:
-            print "vcs err", " ".join(args), "[dir: %s]" % cwd
+            #print "vcs err", " ".join(args), "[dir: %s]" % cwd
             self.log.error(error_msg % self.get_err_msg_ctx(e))
             raise ErrorExit()
         return output.strip()
@@ -367,7 +395,7 @@ class Vcs(object):
     def is_modified(self):
         raise NotImplementedError()
 
-    #m upstream missing?
+    # upstream missing revisions?
 
 
 class Bzr(Vcs):
@@ -387,8 +415,11 @@ class Bzr(Vcs):
         self._call(params, self.err_pull)
 
     def branch(self):
-        params = ["bzr", "branch", self.origin]
-        self._call(params, self.err_branch, os.path.dirname(self.path))
+        params = ["bzr", "branch", self.origin, self.path]
+        cwd = os.path.dirname(os.path.dirname(self.path))
+        if not cwd:
+            cwd = "."
+        self._call(params, self.err_branch, cwd)
 
     def is_modified(self):
         # To replace with bzr cli, we need to be able to detect changes
@@ -397,6 +428,35 @@ class Bzr(Vcs):
         return tree.has_changes()
 
 
+class Git(Vcs):
+
+    def get_cur_rev(self):
+        params = ["git", "rev-parse", "HEAD"]
+        return self._call(params, self.err_cur_rev)
+
+    def update(self,  rev=None):
+        params = ["git", "reset", "--merge"]
+        if rev:
+            params.append(rev)
+        self._call(params, self.err_update)
+
+    def pull(self):
+        params = ["git", "pull", "master"]
+        self._call(params, self.err_pull)
+
+    def branch(self):
+        params = ["git", "clone", self.branch]
+        self._call(params, self.err_branch, os.path.dirname(self.path))
+
+    def is_modified(self):
+        params = ["git", "stat", "-s"]
+        return bool(self._call(params, self.err_is_mod).strip())
+
+    def get_origin(self):
+        params = ["git", "config", "--get", "remote.origin.url"]
+        return self._call(params, "")
+
+
 class Charm(object):
 
     log = logging.getLogger('deployer.charm')
@@ -411,7 +471,12 @@ class Charm(object):
         self.vcs = self.get_vcs()
 
     def get_vcs(self):
-        return Bzr(self.path, self.branch, self.log)
+        if not self.branch:
+            return None
+        if self.branch.startswith('git') or 'github.com' in self.branch:
+            return Git(self.path, self.branch, self.log)
+        elif self.branch.startswith("bzr") or self.branch.startswith('lp:'):
+            return Bzr(self.path, self.branch, self.log)
 
     @classmethod
     def from_service(cls, name, series_path, d):
@@ -440,9 +505,12 @@ class Charm(object):
     def charm_url(self):
         if self._charm_url:
             return self._charm_url
-        return "local:%s" % self.name
+        series = os.path.basename(os.path.dirname(self.path))
+        return "local:%s/%s" % (series, self.name)
 
     def build(self):
+        #if self.charm_url.startswith('cs:'):
+        #    shutil.copy(urllib.urlopen(CHARM_DL_URL % self.charm_url))
         if not self._build:
             return
         self.log.debug("Building charm %s with %s", self.path, self._build)
@@ -488,6 +556,20 @@ class Charm(object):
         with open(md_path) as fh:
             return yaml_load(fh.read())
 
+    def get_provides(self):
+        p = {'juju-info': [{'name': 'juju-info'}]}
+        for key, value in self.metadata['provides'].items():
+            value['name'] = key
+            p.setdefault(value['interface'], []).append(value)
+        return p
+
+    def get_requires(self):
+        r = {}
+        for key, value in self.metadata['requires'].items():
+            value['name'] = key
+            r.setdefault(value['interface'], []).append(value)
+        return r
+
 
 class Deployment(object):
 
@@ -525,10 +607,29 @@ class Deployment(object):
         if 'relations' not in self.data:
             return
 
+        # Strip duplicate rels
+        seen = set()
+
+        def check(a, b):
+            k = tuple(sorted([a, b]))
+            if k in seen:
+                #self.log.warning(" Skipping duplicate relation %r" % (k,))
+                return
+            seen.add(k)
+            return True
+
         # Support an ordered list of [endpoints]
         if isinstance(self.data['relations'], list):
             for end_a, end_b in self.data['relations']:
-                yield (end_a, end_b)
+                if isinstance(end_b, list):
+                    for eb in end_b:
+                        if check(end_a, eb):
+                            yield (end_a, eb)
+                else:
+                    if check(end_a, end_b):
+                        yield (end_a, end_b)
+            #self.log.debug(" Found relations \n  %s" % (
+            #    "\n  ".join(map(str, seen))))
             return
 
         # Legacy format (dictionary of dictionaries with weights)
@@ -540,7 +641,10 @@ class Deployment(object):
             rels[v.get('weight', 0)] = expanded
         for k in sorted(rels):
             for r in rels[k]:
-                yield r
+                if check(*r):
+                    yield r
+
+        #self.log.debug("Found relations %s\n  %s" % (" ".join(map(str, seen))))
 
     def get_charms(self):
         for k, v in self.data['services'].items():
@@ -566,6 +670,13 @@ class Deployment(object):
             else:
                 charm.fetch()
 
+    def resolve(self, cli_overides=()):
+        # Once we have charms definitions available, we can do verification
+        # of config options.
+        self.load_overrides(cli_overides)
+        self.resolve_config()
+        self.validate_relations()
+
     def load_overrides(self, cli_overrides=()):
         """Load overrides."""
         overrides = {}
@@ -620,6 +731,20 @@ class Deployment(object):
                 options[k] = v
             svc_data['options'] = options
 
+    def validate_relations(self):
+        # Could extend to do interface matching against charms.
+        services = dict([(s.name, s) for s in self.get_services()])
+        for e_a, e_b in self.get_relations():
+            for ep in [Endpoint(e_a), Endpoint(e_b)]:
+                if not ep.service in services:
+                    self.log.error(
+                        "Invalid relation in config, service %s not found, rel %s" % (
+                        ep.name, "%s <-> %s" % (e_a, e_b)))
+
+    def save(self, path):
+        with open(path, "w") as fh:
+            fh.write(yaml_dump(self.data, defaul_flow_style=False))
+
 
 class Environment(object):
 
@@ -680,7 +805,7 @@ class Environment(object):
             params.extend(["-e", self.name])
         return params
 
-    def deploy(self, name, charm_url, repo=None, config=None, constraints=None, num_units=1):
+    def deploy(self, name, charm_url, repo=None, config=None, constraints=None, num_units=1, force_machine=None):
         params = self._named_env(["juju", "deploy"])
 
         with temp_file() as fh:
@@ -696,8 +821,12 @@ class Environment(object):
                 if repo == "":
                     repo = "."
                 params.extend(["--repository=%s" % repo])
+            if force_machine is not None:
+                params.extend["--force-machine=%d" % force_machine]
+
             params.extend([charm_url, name])
-            _check_call(params, '', self.log, "Error deploying service %r", name)
+            _check_call(
+                params, '', self.log, "Error deploying service %r", name)
 
     def add_units(self, service_name, num_units):
         return self.client.add_units(service_name, num_units)
@@ -732,7 +861,8 @@ class Environment(object):
 
     def get_cli_status(self):
         params = self._named_env(["juju", "status"])
-        output = _check_call(params, "", self.log, "Error getting status, is it bootstrapped?")
+        output = _check_call(
+            params, "", self.log, "Error getting status, is it bootstrapped?")
         status = yaml_load(output)
         # Opportunistic, see connect method comment.
         if not self.api_endpoint:
@@ -741,20 +871,19 @@ class Environment(object):
         return status
 
     def reset(self, terminate_machines=False, terminate_delay=0, timeout=360, watch=False):
+        """Destroy/reset the environment."""
         status = self.status()
         destroyed = False
         for s in status.get('services', {}).keys():
             self.log.debug(" Destroying service %s", s)
             self.client.destroy_service(s)
             destroyed = True
-            # Check for any units in error and resolve them.
-            for uid, u in status['services'][s].get('units', {}).items():
-                if u['agent-state'] == 'error':
-                    self.log.debug("  Resolving unit error %s for destroy", uid)
-                    self.client.resolved(uid)
 
-        # Wait for units
         if destroyed:
+            # Mark any errors as resolved so destruction can proceed.
+            self.resolve_errors()
+
+            # Wait for units
             self.wait_for_units(timeout, "removed", watch=watch)
 
         # The only value to not terminating is keeping the data on the machines around.
@@ -763,19 +892,70 @@ class Environment(object):
             return
 
         # Terminate machines
-        self.log.info("Terminating machines")
+        self.log.debug(" Terminating machines")
         for mid in status['machines'].keys():
             if mid == "0":
                 continue
-            self.log.debug(" Terminating machine %s", mid)
+            self.log.debug("  Terminating machine %s", mid)
             self.terminate_machine(mid)
             if terminate_delay:
+                self.log.debug("  Waiting for terminate delay")
                 time.sleep(terminate_delay)
 
+    def resolve_errors(self, retry_count=0, timeout=600, watch=False, delay=5):
+        """Resolve any unit errors in the environment.
+
+        If retry_count is given then the hook execution is reattempted. The
+        system will do up to retry_count passes through the system resolving
+        errors.
+
+        If retry count is not given, the error is marked resolved permanently.
+        """
+        etime = time.time() + timeout
+        count = resolved = 0
+        while True:
+            status = self.status()
+            for s in status.get('services', {}).keys():
+                for uid, u in status['services'][s].get('units', {}).items():
+                    if u['agent-state'] == 'error':
+                        self.log.info("  Retrying unit error %s" % (uid))
+                        self.client.resolved(uid, retry=bool(retry_count))
+                        resolved = True
+                    for uid, u in u.get('subordinates', {}).items():
+                        if u['agent-state'] == 'error':
+                            self.log.info("  Retrying subordinate error %s" % (uid))
+                            self.client.resolved(uid, retry=bool(retry_count))
+                            resolved = True
+            if not retry_count or not resolved:
+                self.log.debug("  No errors found.")
+                return
+
+            w_timeout = etime - time.time()
+            if w_timeout < 0:
+                self.log.error("Timeout reached while resolving errors")
+                raise ErrorExit()
+
+            if retry_count:
+                time.sleep(delay)
+
+            count += 1
+            try:
+                self.wait_for_units(
+                    timeout=int(w_timeout), watch=True, no_exit=True)
+            except UnitErrors, e:
+                if retry_count == count:
+                    self.log.info(
+                        " Retry count %d exhausted, but units still in error (%s)",
+                        retry_count, " ".join(u['Name'] for u in e.errors))
+                    return
+            else:
+                return
+
     def status(self):
         return self.client.get_stat()
 
-    def wait_for_units(self, timeout, goal_state="started", watch=False):
+    def wait_for_units(
+            self, timeout, goal_state="started", watch=False, no_exit=False):
         """Wait for units to reach a given condition.
         """
         callback = watch and self._delta_event_log or None
@@ -788,6 +968,8 @@ class Environment(object):
                     u['Name'], u['MachineId'], u['Status'], u['StatusInfo']
                 )
                 for u in e.errors]
+            if no_exit:
+                raise
             self.log.error("The following units had errors:\n   %s" % (
                 "   \n".join(error_units)))
             raise ErrorExit()
@@ -829,14 +1011,18 @@ class Importer(object):
         self.deployment = deployment
 
     def add_units(self):
-        # TODO Check if subordinate
         self.log.debug("Adding units...")
         # Add units to existing services that don't match count.
         env_status = self.env.status()
         added = set()
         for svc in self.deployment.get_services():
-            delta = svc.num_units - len(env_status['services'][svc.name]['units'])
+            delta = svc.num_units - len(env_status['services'][svc.name].get('units', ()))
             if delta > 0:
+                charm = self.deployment.get_charm_for(svc.name)
+                if charm.is_subordinate():
+                    self.log.warning(
+                        "Config specifies num units for subordinate: %s", svc.name)
+                    continue
                 self.log.info(
                     "Adding %d more units to %s" % (abs(delta), svc.name))
                 for u in self.env.add_units(svc.name, abs(delta)):
@@ -851,10 +1037,9 @@ class Importer(object):
         self.deployment.fetch_charms(
             update=self.options.update_charms,
             no_local_mods=self.options.no_local_mods)
-        # Load config overrides after we can validate them.
-        self.deployment.load_overrides(self.options.overrides or ())
-        # Load includes
-        self.deployment.resolve_config()
+
+        # Load config overrides/includes and verify rels after we can validate them.
+        self.deployment.resolve(self.options.overrides or ())
 
     def deploy_services(self):
         self.log.info("Deploying services...")
@@ -873,9 +1058,14 @@ class Importer(object):
                 self.deployment.repo_path,
                 svc.config,
                 svc.constraints,
-                svc.num_units)
+                svc.num_units,
+                svc.force_machine)
+
+            if svc.expose:
+                self.env.expose(svc.name)
 
             if self.options.deploy_delay:
+                self.log.debug(" Waiting for deploy delay")
                 time.sleep(self.options.deploy_delay)
 
     def add_relations(self):
@@ -891,7 +1081,9 @@ class Importer(object):
             self.log.info(" Adding relation %s <-> %s", end_a, end_b)
             self.env.add_relation(end_a, end_b)
             created = True
-            time.sleep(5)  # per the original
+            # per the original, not sure the use case.
+            self.log.debug(" Waiting 5s before next relation")
+            time.sleep(5)
         return created
 
     def _rel_exists(self, status, end_a, end_b):
@@ -917,51 +1109,77 @@ class Importer(object):
             return True
         return False
 
-    def wait_for_units(self):
+    def wait_for_units(self, ignore_error=False):
         timeout = self.options.timeout - (time.time() - self.start_time)
         if timeout < 0:
             self.log.error("Reached deployment timeout.. exiting")
             raise ErrorExit()
-        self.env.wait_for_units(int(timeout), watch=self.options.watch)
+        try:
+            self.env.wait_for_units(
+                int(timeout), watch=self.options.watch, no_exit=ignore_error)
+        except UnitErrors:
+            if not ignore_error:
+                raise
 
     def run(self):
         self.start_time = time.time()
         self.env.connect()
         self.get_charms()
+        if self.options.branch_only:
+            return
+
         self.deploy_services()
         self.wait_for_units()
         self.add_units()
+
         rels_created = self.add_relations()
+
         # Wait for the units to be up before waiting for rel stability.
-        self.wait_for_units()
+        self.wait_for_units(self.options.retry_count)
         if rels_created:
             self.log.debug("Waiting for relations %d", self.options.rel_wait)
             time.sleep(self.options.rel_wait)
 
+        if self.options.retry_count:
+            self.log.info("Looking for errors to auto-retry")
+            self.env.resolve_errors(
+                self.retry_count,
+                self.options.timeout - time.time() - self.start_time)
+
 
 def main():
+    stime = time.time()
+    try:
+        run()
+    except ErrorExit:
+        logging.getLogger('deployer.cli').info(
+            "Deployment stopped. run time: %s", time.time() - stime)
+        sys.exit(1)
+
+
+def run():
     parser = setup_parser()
     (options, args) = parser.parse_args()
-    setup_logging(options)
-    log = logging.getLogger("deployer.cli")
 
+    # Debug implies watching and verbose
+    if options.debug and not options.watch:
+        options.watch = options.verbose = True
+    setup_logging(options.verbose, options.debug)
+
+    log = logging.getLogger("deployer.cli")
     start_time = time.time()
 
     env = Environment(options.juju_env)
     config = ConfigStack(options.configs or [])
 
-    # Debug implies watching
-    if options.debug and not options.watch:
-        options.watch = True
-
     # Destroy services and exit
     if options.destroy_services or options.terminate_machines:
         log.debug("Resetting environment...")
         env.connect()
         env.reset(terminate_machines=options.terminate_machines,
-                  terminate_delay=1,
+                  terminate_delay=options.deploy_delay,
                   watch=options.watch)
-        log.info("Environment reset")
+        log.info("Environment reset in %0.2f", time.time() - start_time)
         sys.exit(0)
 
     # Display service info and exit
@@ -977,6 +1195,13 @@ def main():
             log.info("Service: %s address: %s", options.find_service, address)
         sys.exit(0)
 
+    # Just resolve/retry hooks in the environment
+    if not args and options.retry_count:
+        log.info("Retrying hooks for error resolution")
+        env.connect()
+        env.resolve_errors(
+            options.retry_count, watch=options.watch, timeout=options.timeout)
+
     # Arg check on config files and deployment name.
     if not options.configs:
         log.error("Config files must be specified")
@@ -986,28 +1211,26 @@ def main():
 
     # Just list the available deployments
     if options.list_deploys:
-        log.info("%s", " ".join(config.keys()))
+        log.info("%s", " ".join(sorted(config.keys())))
         sys.exit(0)
 
     # Do something to a deployment
     if not args:
         log.error(
-            "Deployment name must be specified. available: %s", tuple(config.keys()))
+            "Deployment name must be specified. available: %s", tuple(
+                sorted(config.keys())))
         sys.exit(1)
 
     deploy_name = args[0]
     deployment = config.get(deploy_name)
 
-    log.debug("Starting deployment")
-    try:
-        Importer(env, deployment, options).run()
-    except ErrorExit:
-        sys.exit(1)
-    finally:
-        env.close()
+    # Do this thing
+    log.debug("Starting deployment of %s", deploy_name)
+    Importer(env, deployment, options).run()
 
     # Deploy complete
-    log.info("Deployment complete in %0.2f seconds" % (time.time() - start_time))
+    log.info("Deployment complete in %0.2f seconds" % (
+        time.time() - start_time))
 
 
 if __name__ == '__main__':
